@@ -13,25 +13,82 @@
 #include <string.h>
 #include <time.h>
 #include <signal.h>
+#ifdef HAVE_GETOPT_H
+#include <getopt.h>
+#else 
+extern int optind;
+extern char *optarg;
+#endif
+#ifdef HAVE_OPTRESET
+extern int optreset;		/* defined by BSD, but not others */
+#endif
 
 #include "debugfs.h"
 
+/*
+ * This function resets the libc getopt() function, which keeps
+ * internal state.  Bad design!  Stupid libc API designers!  No
+ * biscuit!
+ *
+ * BSD-derived getopt() functions require that optind be reset to 1 in
+ * order to reset getopt() state.  This used to be generally accepted
+ * way of resetting getopt().  However, glibc's getopt()
+ * has additional getopt() state beyond optind, and requires that
+ * optind be set zero to reset its state.  So the unfortunate state of
+ * affairs is that BSD-derived versions of getopt() misbehave if
+ * optind is set to 0 in order to reset getopt(), and glibc's getopt()
+ * will core dump if optind is set 1 in order to reset getopt().
+ * 
+ * More modern versions of BSD require that optreset be set to 1 in
+ * order to reset getopt().   Sigh.  Standards, anyone?
+ *
+ * We hide the hair here.
+ */
+void reset_getopt(void)
+{
+#if defined(__GLIBC__) || defined(__linux__)
+	optind = 0;
+#else
+	optind = 1;
+#endif
+#ifdef HAVE_OPTRESET
+	optreset = 1;		/* Makes BSD getopt happy */
+#endif
+}
+
+static const char *pager_search_list[] = { "pager", "more", "less", 0 };
+static const char *pager_dir_list[] = { "/usr/bin", "/bin", 0 };
+
+static const char *find_pager(char *buf)
+{
+	const char **i, **j;
+
+	for (i = pager_search_list; *i; i++) {
+		for (j = pager_dir_list; *j; j++) {
+			sprintf(buf, "%s/%s", *j, *i);
+			if (access(buf, X_OK) == 0)
+				return(buf);
+		}
+	}
+	return 0;
+}
+
 FILE *open_pager(void)
 {
-	FILE *outfile;
-	const char *pager = getenv("PAGER");
+	FILE *outfile = 0;
+	const char *pager = getenv("DEBUGFS_PAGER");
+	char buf[80];
 
 	signal(SIGPIPE, SIG_IGN);
-	if (pager) {
-		if (strcmp(pager, "__none__") == 0) {
-			return stdout;
-		}
-	} else
-		pager = "more";
-
-	outfile = popen(pager, "w");
-
-	return (outfile ? outfile : stdout);
+	if (!pager)
+		pager = getenv("PAGER");
+	if (!pager)
+		pager = find_pager(buf);
+	if (!pager || 
+	    (strcmp(pager, "__none__") == 0) ||
+	    ((outfile = popen(pager, "w")) == 0))
+		return stdout;
+	return outfile;
 }
 
 void close_pager(FILE *stream)
@@ -129,9 +186,55 @@ int check_fs_bitmaps(char *name)
  */
 char *time_to_string(__u32 cl)
 {
-	time_t	t = (time_t) cl;
+	static int	do_gmt = -1;
+	time_t		t = (time_t) cl;
+	char *		tz;
 
-	return ctime(&t);
+	if (do_gmt == -1) {
+		/* The diet libc doesn't respect the TZ environemnt variable */
+		tz = getenv("TZ");
+		if (!tz)
+			tz = "";
+		do_gmt = !strcmp(tz, "GMT");
+	}
+
+	return asctime((do_gmt) ? gmtime(&t) : localtime(&t));
+}
+
+/*
+ * Parse a string as a time.  Return ((time_t)-1) if the string
+ * doesn't appear to be a sane time.
+ */
+extern time_t string_to_time(const char *arg)
+{
+	struct	tm	ts;
+	unsigned long	ret;
+	char *tmp;
+
+	if (strcmp(arg, "now") == 0) {
+		return time(0);
+	}
+	memset(&ts, 0, sizeof(ts));
+#ifdef HAVE_STRPTIME
+	strptime(arg, "%Y%m%d%H%M%S", &ts);
+#else
+	sscanf(arg, "%4d%2d%2d%2d%2d%2d", &ts.tm_year, &ts.tm_mon,
+	       &ts.tm_mday, &ts.tm_hour, &ts.tm_min, &ts.tm_sec);
+	ts.tm_year -= 1900;
+	ts.tm_mon -= 1;
+	if (ts.tm_year < 0 || ts.tm_mon < 0 || ts.tm_mon > 11 ||
+	    ts.tm_mday < 0 || ts.tm_mday > 31 || ts.tm_hour > 23 ||
+	    ts.tm_min > 59 || ts.tm_sec > 61)
+		ts.tm_mday = 0;
+#endif
+	if (ts.tm_mday == 0) {
+		/* Try it as an integer... */
+
+		ret = strtoul(arg, &tmp, 0);
+		if (*tmp)
+			return ((time_t) -1);
+	}
+	return mktime(&ts);
 }
 
 /*
@@ -240,6 +343,19 @@ int common_block_args_process(int argc, char *argv[],
 	return 0;
 }
 
+int debugfs_read_inode_full(ext2_ino_t ino, struct ext2_inode * inode,
+			const char *cmd, int bufsize)
+{
+	int retval;
+
+	retval = ext2fs_read_inode_full(current_fs, ino, inode, bufsize);
+	if (retval) {
+		com_err(cmd, retval, "while reading inode %u", ino);
+		return 1;
+	}
+	return 0;
+}
+
 int debugfs_read_inode(ext2_ino_t ino, struct ext2_inode * inode,
 			const char *cmd)
 {
@@ -266,3 +382,15 @@ int debugfs_write_inode(ext2_ino_t ino, struct ext2_inode * inode,
 	return 0;
 }
 
+int debugfs_write_new_inode(ext2_ino_t ino, struct ext2_inode * inode,
+			    const char *cmd)
+{
+	int retval;
+
+	retval = ext2fs_write_new_inode(current_fs, ino, inode);
+	if (retval) {
+		com_err(cmd, retval, "while creating inode %u", ino);
+		return 1;
+	}
+	return 0;
+}
