@@ -298,7 +298,7 @@ static int release_orphan_inodes(e2fsck_t ctx)
 		if (!inode.i_links_count) {
 			ext2fs_inode_alloc_stats2(fs, ino, -1,
 						  LINUX_S_ISDIR(inode.i_mode));
-			inode.i_dtime = time(0);
+			inode.i_dtime = ctx->now;
 		} else {
 			inode.i_dtime = 0;
 		}
@@ -346,7 +346,7 @@ void check_resize_inode(e2fsck_t ctx)
 		}
 	}
 
-	/* Read the resizde inode */
+	/* Read the resize inode */
 	pctx.ino = EXT2_RESIZE_INO;
 	retval = ext2fs_read_inode(fs, EXT2_RESIZE_INO, &inode);
 	if (retval) {
@@ -705,9 +705,34 @@ void check_super_block(e2fsck_t ctx)
 		ext2fs_mark_super_dirty(fs);
 	}
 
+	/* 
+	 * Check to see if the superblock last mount time or last
+	 * write time is in the future.
+	 */
+	if (fs->super->s_mtime > ctx->now) {
+		pctx.num = fs->super->s_mtime;
+		if (fix_problem(ctx, PR_0_FUTURE_SB_LAST_MOUNT, &pctx)) {
+			fs->super->s_mtime = ctx->now;
+			ext2fs_mark_super_dirty(fs);
+		}
+	}
+	if (fs->super->s_wtime > ctx->now) {
+		pctx.num = fs->super->s_wtime;
+		if (fix_problem(ctx, PR_0_FUTURE_SB_LAST_WRITE, &pctx)) {
+			fs->super->s_wtime = ctx->now;
+			ext2fs_mark_super_dirty(fs);
+		}
+	}
+
 	/*
 	 * Move the ext3 journal file, if necessary.
 	 */
 	e2fsck_move_ext3_journal(ctx);
+
+	/*
+	 * Fix journal hint, if necessary
+	 */
+	e2fsck_fix_ext3_journal_hint(ctx);
+
 	return;
 }
