@@ -60,6 +60,7 @@ extern "C" {
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 
 #if EXT2_FLAT_INCLUDES
 #include "e2_types.h"
@@ -187,6 +188,7 @@ typedef struct ext2_file *ext2_file_t;
 #define EXT2_FLAG_SUPER_ONLY		0x800
 #define EXT2_FLAG_JOURNAL_DEV_OK	0x1000
 #define EXT2_FLAG_IMAGE_FILE		0x2000
+#define EXT2_FLAG_EXCLUSIVE		0x4000
 
 /*
  * Special flag in the ext2 inode i_flag field that means that this is
@@ -228,10 +230,11 @@ struct struct_ext2_filsys {
 	struct ext2_super_block *	orig_super;
 	struct ext2_image_hdr *		image_header;
 	__u32				umask;
+	time_t				now;
 	/*
 	 * Reserved for future expansion
 	 */
-	__u32				reserved[8];
+	__u32				reserved[7];
 
 	/*
 	 * Reserved for the use of the calling application.
@@ -267,9 +270,9 @@ struct struct_ext2_filsys {
  * to an inode.  It can also be used for programs that want to be able
  * to deal with files that contain "holes".
  * 
- * BLOCK_FLAG_TRAVERSE indicates that the iterator function for the
- * indirect, doubly indirect, etc. blocks should be called after all
- * of the blocks containined in the indirect blocks are processed.
+ * BLOCK_FLAG_DEPTH_TRAVERSE indicates that the iterator function for
+ * the indirect, doubly indirect, etc. blocks should be called after
+ * all of the blocks containined in the indirect blocks are processed.
  * This is useful if you are going to be deallocating blocks from an
  * inode.
  *
@@ -346,6 +349,7 @@ typedef struct ext2_struct_inode_scan *ext2_inode_scan;
 #define EXT2_MF_ISROOT		2
 #define EXT2_MF_READONLY	4
 #define EXT2_MF_SWAP		8
+#define EXT2_MF_BUSY		16
 
 /*
  * Ext2/linux mode flags.  We define them here so that we don't need
@@ -982,6 +986,7 @@ extern blk_t ext2fs_inode_data_blocks(ext2_filsys fs,
 #endif
 
 #ifndef EXT2_CUSTOM_MEMORY_ROUTINES
+#include <string.h>
 /*
  *  Allocate memory
  */
@@ -1014,12 +1019,14 @@ _INLINE_ errcode_t ext2fs_resize_mem(unsigned long EXT2FS_ATTR((unused)) old_siz
 				     unsigned long size, void *ptr)
 {
 	void *p;
-	void **pp = (void **)ptr;
 
-	p = realloc(*pp, size);
+	/* Use "memcpy" for pointer assignments here to avoid problems
+	 * with C99 strict type aliasing rules. */
+	memcpy(&p, ptr, sizeof (p));
+	p = realloc(p, size);
 	if (!p)
 		return EXT2_ET_NO_MEMORY;
-	*pp = p;
+	memcpy(ptr, &p, sizeof (p));
 	return 0;
 }
 #endif	/* Custom memory routines */
