@@ -23,42 +23,6 @@
 
 #include "ext2fs.h"
 
-#ifdef EXT2FS_ENABLE_SWAPFS
-void ext2fs_swap_ext_attr(ext2_filsys fs, char *to, char *from)
-{
-	struct ext2_ext_attr_header *from_header =
-		(struct ext2_ext_attr_header *)from;
-	struct ext2_ext_attr_header *to_header =
-		(struct ext2_ext_attr_header *)to;
-	struct ext2_ext_attr_entry *from_entry, *to_entry;
-	char *from_end = (char *)from_header + fs->blocksize;
-	int n;
-
-	if (to_header != from_header)
-		memcpy(to_header, from_header, fs->blocksize);
-
-	to_header->h_magic    = ext2fs_swab32(from_header->h_magic);
-	to_header->h_blocks   = ext2fs_swab32(from_header->h_blocks);
-	to_header->h_refcount = ext2fs_swab32(from_header->h_refcount);
-	for (n=0; n<4; n++)
-		to_header->h_reserved[n] =
-			ext2fs_swab32(from_header->h_reserved[n]);
-	
-	from_entry = (struct ext2_ext_attr_entry *)(from_header+1);
-	to_entry   = (struct ext2_ext_attr_entry *)(to_header+1);
-	while ((char *)from_entry < from_end && *(__u32 *)from_entry) {
-		to_entry->e_value_offs  =	
-			ext2fs_swab16(from_entry->e_value_offs);
-		to_entry->e_value_block =	
-			ext2fs_swab32(from_entry->e_value_block);
-		to_entry->e_value_size  =	
-			ext2fs_swab32(from_entry->e_value_size);
-		from_entry = EXT2_EXT_ATTR_NEXT(from_entry);
-		to_entry   = EXT2_EXT_ATTR_NEXT(to_entry);
-	}
-}
-#endif
-
 errcode_t ext2fs_read_ext_attr(ext2_filsys fs, blk_t block, void *buf)
 {
 	errcode_t	retval;
@@ -69,7 +33,7 @@ errcode_t ext2fs_read_ext_attr(ext2_filsys fs, blk_t block, void *buf)
 #ifdef EXT2FS_ENABLE_SWAPFS
 	if ((fs->flags & (EXT2_FLAG_SWAP_BYTES|
 			  EXT2_FLAG_SWAP_BYTES_READ)) != 0)
-		ext2fs_swap_ext_attr(fs, buf, buf);
+		ext2fs_swap_ext_attr(buf, buf, fs->blocksize, 1);
 #endif
 	return 0;
 }
@@ -83,17 +47,17 @@ errcode_t ext2fs_write_ext_attr(ext2_filsys fs, blk_t block, void *inbuf)
 #ifdef EXT2FS_ENABLE_SWAPFS
 	if ((fs->flags & EXT2_FLAG_SWAP_BYTES) ||
 	    (fs->flags & EXT2_FLAG_SWAP_BYTES_WRITE)) {
-		retval = ext2fs_get_mem(fs->blocksize, (void **) &buf);
+		retval = ext2fs_get_mem(fs->blocksize, &buf);
 		if (retval)
 			return retval;
 		write_buf = buf;
-		ext2fs_swap_ext_attr(fs, buf, inbuf);
+		ext2fs_swap_ext_attr(buf, inbuf, fs->blocksize, 1);
 	} else
 #endif
 		write_buf = (char *) inbuf;
  	retval = io_channel_write_blk(fs->io, block, 1, write_buf);
 	if (buf)
-		ext2fs_free_mem((void **) &buf);
+		ext2fs_free_mem(&buf);
 	if (!retval)
 		ext2fs_mark_changed(fs);
 	return retval;
@@ -115,7 +79,7 @@ errcode_t ext2fs_adjust_ea_refcount(ext2_filsys fs, blk_t blk,
 		return EXT2_ET_BAD_EA_BLOCK_NUM;
 
 	if (!block_buf) {
-		retval = ext2fs_get_mem(fs->blocksize, (void **) &buf);
+		retval = ext2fs_get_mem(fs->blocksize, &buf);
 		if (retval)
 			return retval;
 		block_buf = buf;
@@ -136,6 +100,6 @@ errcode_t ext2fs_adjust_ea_refcount(ext2_filsys fs, blk_t blk,
 
 errout:
 	if (buf)
-		ext2fs_free_mem((void **) &buf);
+		ext2fs_free_mem(&buf);
 	return retval;
 }
