@@ -51,7 +51,7 @@ errcode_t ext2fs_create_journal_superblock(ext2_filsys fs,
 	if (size < 1024)
 		return EXT2_ET_JOURNAL_TOO_SMALL;
 
-	if ((retval = ext2fs_get_mem(fs->blocksize, (void **) &jsb)))
+	if ((retval = ext2fs_get_mem(fs->blocksize, &jsb)))
 		return retval;
 
 	memset (jsb, 0, fs->blocksize);
@@ -94,7 +94,8 @@ static errcode_t write_journal_file(ext2_filsys fs, char *filename,
 {
 	errcode_t	retval;
 	char		*buf = 0;
-	int		i, fd, ret_size;
+	int		fd, ret_size;
+	blk_t		i;
 
 	if ((retval = ext2fs_create_journal_superblock(fs, size, flags, &buf)))
 		return retval;
@@ -112,7 +113,7 @@ static errcode_t write_journal_file(ext2_filsys fs, char *filename,
 		retval = errno;
 		goto errout;
 	}
-	if (ret_size != fs->blocksize)
+	if (ret_size != (int) fs->blocksize)
 		goto errout;
 	memset(buf, 0, fs->blocksize);
 
@@ -122,14 +123,14 @@ static errcode_t write_journal_file(ext2_filsys fs, char *filename,
 			retval = errno;
 			goto errout;
 		}
-		if (ret_size != fs->blocksize)
+		if (ret_size != (int) fs->blocksize)
 			goto errout;
 	}
 	close(fd);
 
 	retval = 0;
 errout:
-	ext2fs_free_mem((void **) &buf);
+	ext2fs_free_mem(&buf);
 	return retval;
 }
 
@@ -143,12 +144,12 @@ struct mkjournal_struct {
 	errcode_t	err;
 };
 
-static int mkjournal_proc(ext2_filsys		fs,
-			   blk_t		*blocknr,
-			   e2_blkcnt_t		blockcnt,
-			   blk_t		ref_block,
-			   int			ref_offset,
-			   void			*priv_data)
+static int mkjournal_proc(ext2_filsys	fs,
+			   blk_t	*blocknr,
+			   e2_blkcnt_t	blockcnt,
+			   blk_t	ref_block EXT2FS_ATTR((unused)),
+			   int		ref_offset EXT2FS_ATTR((unused)),
+			   void		*priv_data)
 {
 	struct mkjournal_struct *es = (struct mkjournal_struct *) priv_data;
 	blk_t	new_blk;
@@ -228,7 +229,7 @@ static errcode_t write_journal_inode(ext2_filsys fs, ext2_ino_t journal_ino,
 
  	inode.i_size += fs->blocksize * size;
 	inode.i_blocks += (fs->blocksize / 512) * es.newblocks;
-	inode.i_mtime = inode.i_ctime = time(0);
+	inode.i_mtime = inode.i_ctime = fs->now ? fs->now : time(0);
 	inode.i_links_count = 1;
 	inode.i_mode = LINUX_S_IFREG | 0600;
 
@@ -236,8 +237,13 @@ static errcode_t write_journal_inode(ext2_filsys fs, ext2_ino_t journal_ino,
 		goto errout;
 	retval = 0;
 
+	memcpy(fs->super->s_jnl_blocks, inode.i_block, EXT2_N_BLOCKS*4);
+	fs->super->s_jnl_blocks[16] = inode.i_size;
+	fs->super->s_jnl_backup_type = EXT3_JNL_BACKUP_BLOCKS;
+	ext2fs_mark_super_dirty(fs);
+
 errout:
-	ext2fs_free_mem((void **) &buf);
+	ext2fs_free_mem(&buf);
 	return retval;
 }
 
@@ -250,8 +256,8 @@ errcode_t ext2fs_add_journal_device(ext2_filsys fs, ext2_filsys journal_dev)
 	errcode_t	retval;
 	char		buf[1024];
 	journal_superblock_t	*jsb;
-	int		i, start;
-	__u32		nr_users;
+	int		start;
+	__u32		i, nr_users;
 
 	/* Make sure the device exists and is a block device */
 	if (stat(journal_dev->device_name, &st) < 0)
@@ -272,7 +278,7 @@ errcode_t ext2fs_add_journal_device(ext2_filsys fs, ext2_filsys journal_dev)
 	    (jsb->s_header.h_blocktype != (unsigned) ntohl(JFS_SUPERBLOCK_V2)))
 		return EXT2_ET_NO_JOURNAL_SB;
 
-	if (ntohl(jsb->s_blocksize) != fs->blocksize)
+	if (ntohl(jsb->s_blocksize) != (unsigned long) fs->blocksize)
 		return EXT2_ET_UNEXPECTED_BLOCK_SIZE;
 
 	/* Check and see if this filesystem has already been added */
@@ -312,7 +318,8 @@ errcode_t ext2fs_add_journal_inode(ext2_filsys fs, blk_t size, int flags)
 	ext2_ino_t		journal_ino;
 	struct stat		st;
 	char			jfile[1024];
-	int			fd, mount_flags, f;
+	int			mount_flags, f;
+	int			fd = -1;
 
 	if ((retval = ext2fs_check_mount_point(fs->device_name, &mount_flags,
 					       jfile, sizeof(jfile)-10)))
@@ -363,6 +370,10 @@ errcode_t ext2fs_add_journal_inode(ext2_filsys fs, blk_t size, int flags)
 		close(fd);
 		journal_ino = st.st_ino;
 	} else {
+		if (mount_flags & EXT2_MF_BUSY) {
+			retval = EBUSY;
+			goto errout;
+		}
 		journal_ino = EXT2_JOURNAL_INO;
 		if ((retval = write_journal_inode(fs, journal_ino,
 						  size, flags)))
@@ -378,7 +389,8 @@ errcode_t ext2fs_add_journal_inode(ext2_filsys fs, blk_t size, int flags)
 	ext2fs_mark_super_dirty(fs);
 	return 0;
 errout:
-	close(fd);
+	if (fd > 0)
+		close(fd);
 	return retval;
 }
 
