@@ -55,8 +55,8 @@ struct unix_cache {
 	char		*buf;
 	unsigned long	block;
 	int		access_time;
-	int		dirty:1;
-	int		in_use:1;
+	unsigned	dirty:1;
+	unsigned	in_use:1;
 };
 
 #define CACHE_SIZE 8
@@ -68,6 +68,7 @@ struct unix_private_data {
 	int	dev;
 	int	flags;
 	int	access_time;
+	ext2_loff_t offset;
 	struct unix_cache cache[CACHE_SIZE];
 };
 
@@ -81,10 +82,20 @@ static errcode_t unix_write_blk(io_channel channel, unsigned long block,
 static errcode_t unix_flush(io_channel channel);
 static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
 				int size, const void *data);
+static errcode_t unix_set_option(io_channel channel, const char *option, 
+				 const char *arg);
 
 static void reuse_cache(io_channel channel, struct unix_private_data *data,
 		 struct unix_cache *cache, unsigned long block);
 
+/* __FreeBSD_kernel__ is defined by GNU/kFreeBSD - the FreeBSD kernel
+ * does not know buffered block devices - everything is raw. */
+#if defined(__CYGWIN__) || defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
+#define NEED_BOUNCE_BUFFER
+#else
+#undef NEED_BOUNCE_BUFFER
+#endif
+
 static struct struct_io_manager struct_unix_manager = {
 	EXT2_ET_MAGIC_IO_MANAGER,
 	"Unix I/O Manager",
@@ -94,11 +105,12 @@ static struct struct_io_manager struct_unix_manager = {
 	unix_read_blk,
 	unix_write_blk,
 	unix_flush,
-#ifdef __CYGWIN__
-	0
+#ifdef NEED_BOUNCE_BUFFER
+	0,
 #else
-	unix_write_byte
+	unix_write_byte,
 #endif
+	unix_set_option
 };
 
 io_manager unix_io_manager = &struct_unix_manager;
@@ -106,19 +118,19 @@ io_manager unix_io_manager = &struct_unix_manager;
 /*
  * Here are the raw I/O functions
  */
-#ifndef __CYGWIN__
+#ifndef NEED_BOUNCE_BUFFER
 static errcode_t raw_read_blk(io_channel channel,
 			      struct unix_private_data *data,
 			      unsigned long block,
 			      int count, void *buf)
 {
 	errcode_t	retval;
-	size_t		size;
+	ssize_t		size;
 	ext2_loff_t	location;
 	int		actual = 0;
 
 	size = (count < 0) ? -count : count * channel->block_size;
-	location = (ext2_loff_t) block * channel->block_size;
+	location = ((ext2_loff_t) block * channel->block_size) + data->offset;
 	if (ext2fs_llseek(data->dev, location, SEEK_SET) != location) {
 		retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
 		goto error_out;
@@ -139,9 +151,9 @@ error_out:
 					       size, actual, retval);
 	return retval;
 }
-#else /* __CYGWIN__ */
+#else /* NEED_BOUNCE_BUFFER */
 /*
- * Windows block devices only allow sector alignment IO in offset and size
+ * Windows and FreeBSD block devices only allow sector alignment IO in offset and size
  */
 static errcode_t raw_read_blk(io_channel channel,
 			      struct unix_private_data *data,
@@ -156,7 +168,7 @@ static errcode_t raw_read_blk(io_channel channel,
 	char		sector[BLOCKALIGN];
 
 	size = (count < 0) ? -count : count * channel->block_size;
-	location = (ext2_loff_t) block * channel->block_size;
+	location = ((ext2_loff_t) block * channel->block_size) + data->offset;
 #ifdef DEBUG
 	printf("count=%d, size=%d, block=%d, blk_size=%d, location=%lx\n",
 	 		count, size, block, channel->block_size, location);
@@ -199,7 +211,7 @@ static errcode_t raw_write_blk(io_channel channel,
 			       unsigned long block,
 			       int count, const void *buf)
 {
-	size_t		size;
+	ssize_t		size;
 	ext2_loff_t	location;
 	int		actual = 0;
 	errcode_t	retval;
@@ -213,7 +225,7 @@ static errcode_t raw_write_blk(io_channel channel,
 			size = count * channel->block_size;
 	}
 
-	location = (ext2_loff_t) block * channel->block_size;
+	location = ((ext2_loff_t) block * channel->block_size) + data->offset;
 	if (ext2fs_llseek(data->dev, location, SEEK_SET) != location) {
 		retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
 		goto error_out;
@@ -253,15 +265,14 @@ static errcode_t alloc_cache(io_channel channel,
 		cache->dirty = 0;
 		cache->in_use = 0;
 		if ((retval = ext2fs_get_mem(channel->block_size,
-					     (void **) &cache->buf)))
+					     &cache->buf)))
 			return retval;
 	}
 	return 0;
 }
 
 /* Free the cache buffers */
-static void free_cache(io_channel channel, 
-		       struct unix_private_data *data)
+static void free_cache(struct unix_private_data *data)
 {
 	struct unix_cache	*cache;
 	int			i;
@@ -273,7 +284,7 @@ static void free_cache(io_channel channel,
 		cache->dirty = 0;
 		cache->in_use = 0;
 		if (cache->buf)
-			ext2fs_free_mem((void **) &cache->buf);
+			ext2fs_free_mem(&cache->buf);
 		cache->buf = 0;
 	}
 }
@@ -284,8 +295,7 @@ static void free_cache(io_channel channel,
  * eldest is a non-zero pointer, then fill in eldest with the cache
  * entry to that should be reused.
  */
-static struct unix_cache *find_cached_block(io_channel channel,
-					    struct unix_private_data *data,
+static struct unix_cache *find_cached_block(struct unix_private_data *data,
 					    unsigned long block,
 					    struct unix_cache **eldest)
 {
@@ -374,19 +384,17 @@ static errcode_t unix_open(const char *name, int flags, io_channel *channel)
 
 	if (name == 0)
 		return EXT2_ET_BAD_DEVICE_NAME;
-	retval = ext2fs_get_mem(sizeof(struct struct_io_channel),
-				(void **) &io);
+	retval = ext2fs_get_mem(sizeof(struct struct_io_channel), &io);
 	if (retval)
 		return retval;
 	memset(io, 0, sizeof(struct struct_io_channel));
 	io->magic = EXT2_ET_MAGIC_IO_CHANNEL;
-	retval = ext2fs_get_mem(sizeof(struct unix_private_data),
-				(void **) &data);
+	retval = ext2fs_get_mem(sizeof(struct unix_private_data), &data);
 	if (retval)
 		goto cleanup;
 
 	io->manager = unix_io_manager;
-	retval = ext2fs_get_mem(strlen(name)+1, (void **) &io->name);
+	retval = ext2fs_get_mem(strlen(name)+1, &io->name);
 	if (retval)
 		goto cleanup;
 
@@ -402,12 +410,14 @@ static errcode_t unix_open(const char *name, int flags, io_channel *channel)
 
 	if ((retval = alloc_cache(io, data)))
 		goto cleanup;
-	
+
 	open_flags = (flags & IO_FLAG_RW) ? O_RDWR : O_RDONLY;
+	if (flags & IO_FLAG_EXCLUSIVE)
+		open_flags |= O_EXCL;
 #ifdef HAVE_OPEN64
-	data->dev = open64(name, open_flags);
+	data->dev = open64(io->name, open_flags);
 #else
-	data->dev = open(name, open_flags);
+	data->dev = open(io->name, open_flags);
 #endif
 	if (data->dev < 0) {
 		retval = errno;
@@ -453,11 +463,11 @@ static errcode_t unix_open(const char *name, int flags, io_channel *channel)
 
 cleanup:
 	if (data) {
-		free_cache(io, data);
-		ext2fs_free_mem((void **) &data);
+		free_cache(data);
+		ext2fs_free_mem(&data);
 	}
 	if (io)
-		ext2fs_free_mem((void **) &io);
+		ext2fs_free_mem(&io);
 	return retval;
 }
 
@@ -479,12 +489,12 @@ static errcode_t unix_close(io_channel channel)
 
 	if (close(data->dev) < 0)
 		retval = errno;
-	free_cache(channel, data);
+	free_cache(data);
 
-	ext2fs_free_mem((void **) &channel->private_data);
+	ext2fs_free_mem(&channel->private_data);
 	if (channel->name)
-		ext2fs_free_mem((void **) &channel->name);
-	ext2fs_free_mem((void **) &channel);
+		ext2fs_free_mem(&channel->name);
+	ext2fs_free_mem(&channel);
 	return retval;
 }
 
@@ -504,7 +514,7 @@ static errcode_t unix_set_blksize(io_channel channel, int blksize)
 #endif
 		
 		channel->block_size = blksize;
-		free_cache(channel, data);
+		free_cache(data);
 		if ((retval = alloc_cache(channel, data)))
 			return retval;
 	}
@@ -541,8 +551,7 @@ static errcode_t unix_read_blk(io_channel channel, unsigned long block,
 	cp = buf;
 	while (count > 0) {
 		/* If it's in the cache, use it! */
-		if ((cache = find_cached_block(channel, data, block,
-					       &reuse[0]))) {
+		if ((cache = find_cached_block(data, block, &reuse[0]))) {
 #ifdef DEBUG
 			printf("Using cached block %d\n", block);
 #endif
@@ -557,8 +566,7 @@ static errcode_t unix_read_blk(io_channel channel, unsigned long block,
 		 * single read request
 		 */
 		for (i=1; i < count; i++)
-			if (find_cached_block(channel, data, block+i,
-					      &reuse[i]))
+			if (find_cached_block(data, block+i, &reuse[i]))
 				break;
 #ifdef DEBUG
 		printf("Reading %d blocks starting at %d\n", i, block);
@@ -616,7 +624,7 @@ static errcode_t unix_write_blk(io_channel channel, unsigned long block,
 	
 	cp = buf;
 	while (count > 0) {
-		cache = find_cached_block(channel, data, block, &reuse);
+		cache = find_cached_block(data, block, &reuse);
 		if (!cache) {
 			cache = reuse;
 			reuse_cache(channel, data, cache, block);
@@ -636,7 +644,7 @@ static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
 {
 	struct unix_private_data *data;
 	errcode_t	retval = 0;
-	size_t		actual;
+	ssize_t		actual;
 
 	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
 	data = (struct unix_private_data *) channel->private_data;
@@ -650,7 +658,7 @@ static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
 		return retval;
 #endif
 
-	if (lseek(data->dev, offset, SEEK_SET) < 0)
+	if (lseek(data->dev, offset + data->offset, SEEK_SET) < 0)
 		return errno;
 	
 	actual = write(data->dev, buf, size);
@@ -679,3 +687,26 @@ static errcode_t unix_flush(io_channel channel)
 	return retval;
 }
 
+static errcode_t unix_set_option(io_channel channel, const char *option, 
+				 const char *arg)
+{
+	struct unix_private_data *data;
+	unsigned long tmp;
+	char *end;
+
+	EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
+	data = (struct unix_private_data *) channel->private_data;
+	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);
+
+	if (!strcmp(option, "offset")) {
+		if (!arg)
+			return EXT2_ET_INVALID_ARGUMENT;
+
+		tmp = strtoul(arg, &end, 0);
+		if (*end)
+			return EXT2_ET_INVALID_ARGUMENT;
+		data->offset = tmp;
+		return 0;
+	}
+	return EXT2_ET_INVALID_ARGUMENT;
+}
