@@ -57,7 +57,7 @@ extern int optind;
 #include "nls-enable.h"
 
 const char * program_name = "badblocks";
-const char * done_string = N_("done                        \n");
+const char * done_string = N_("done                             \n");
 
 static int v_flag = 0;			/* verbose */
 static int w_flag = 0;			/* do r/w test: 0=no, 1=yes,
@@ -780,17 +780,27 @@ static void check_mount(char *device_name)
 			device_name);
 		return;
 	}
-	if (!(mount_flags & EXT2_MF_MOUNTED))
-		return;
-
-	fprintf(stderr, _("%s is mounted; "), device_name);
-	if (force) {
-		fputs(_("badblocks forced anyway.  "
-			"Hope /etc/mtab is incorrect.\n"), stderr);
-		return;
+	if (mount_flags & EXT2_MF_MOUNTED) {
+		fprintf(stderr, _("%s is mounted; "), device_name);
+		if (force) {
+			fputs(_("badblocks forced anyway.  "
+				"Hope /etc/mtab is incorrect.\n"), stderr);
+			return;
+		}
+	abort_badblocks:
+		fputs(_("it's not safe to run badblocks!\n"), stderr);
+		exit(1);
 	}
-	fputs(_("it's not safe to run badblocks!\n"), stderr);
-	exit(1);
+
+	if (mount_flags & EXT2_MF_BUSY) {
+		fprintf(stderr, _("%s is apparently in use by the system; "),
+			device_name);
+		if (force)
+			fputs(_("badblocks forced anyway.\n"), stderr);
+		else
+			goto abort_badblocks;
+	}
+
 }
 
 
@@ -967,7 +977,7 @@ int main (int argc, char ** argv)
 	} else {
 		last_block = strtoul (argv[optind], &tmp, 0);
 		if (*tmp) {
-			com_err (program_name, 0, _("bad blocks count - %s"),
+			com_err (program_name, 0, _("invalid blocks count - %s"),
 				 argv[optind]);
 			exit (1);
 		}
@@ -976,13 +986,13 @@ int main (int argc, char ** argv)
 	if (optind <= argc-1) {
 		from_count = strtoul (argv[optind], &tmp, 0);
 		if (*tmp) {
-			com_err (program_name, 0, _("bad starting block - %s"),
+			com_err (program_name, 0, _("invalid starting block - %s"),
 				 argv[optind]);
 			exit (1);
 		}
 	} else from_count = 0;
 	if (from_count >= last_block) {
-	    com_err (program_name, 0, _("bad blocks range: %lu-%lu"),
+	    com_err (program_name, 0, _("invalid blocks range: %lu-%lu"),
 		     (unsigned long) from_count, (unsigned long) last_block);
 	    exit (1);
 	}
@@ -1037,7 +1047,7 @@ int main (int argc, char ** argv)
 	errcode = ext2fs_badblocks_list_create(&bb_list,0);
 	if (errcode) {
 		com_err (program_name, errcode,
-			 _("creating in-memory bad blocks list"));
+			 _("while creating in-memory bad blocks list"));
 		exit (1);
 	}
 
@@ -1052,7 +1062,7 @@ int main (int argc, char ** argv)
 				default:
 					errcode = ext2fs_badblocks_list_add(bb_list,next_bad);
 					if (errcode) {
-						com_err (program_name, errcode, _("adding to in-memory bad block list"));
+						com_err (program_name, errcode, _("while adding to in-memory bad block list"));
 						exit (1);
 					}
 					continue;
