@@ -64,7 +64,8 @@ static int m_flag, M_flag, r_flag, s_flag = -1, u_flag, U_flag, T_flag;
 static time_t last_check_time;
 static int print_label;
 static int max_mount_count, mount_count, mount_flags;
-static unsigned long interval, reserved_ratio, reserved_blocks;
+static unsigned long interval, reserved_blocks;
+static double reserved_ratio;
 static unsigned long resgid, resuid;
 static unsigned short errors;
 static int open_flag;
@@ -81,14 +82,14 @@ void do_findfs(int argc, char **argv);
 static void usage(void)
 {
 	fprintf(stderr,
-		_("Usage: %s [-c max-mounts-count] [-e errors-behavior] "
+		_("Usage: %s [-c max_mounts_count] [-e errors_behavior] "
 		  "[-g group]\n"
-		  "\t[-i interval[d|m|w]] [-j] [-J journal-options]\n"
-		  "\t[-l] [-s sparse-flag] [-m reserved-blocks-percent]\n"
-		  "\t[-o [^]mount-options[,...]] [-r reserved-blocks-count]\n"
-		  "\t[-u user] [-C mount-count] [-L volume-label] "
-		  "[-M last-mounted-dir]\n"
-		  "\t[-O [^]feature[,...]] [-T last-check-time] [-U UUID]"
+		  "\t[-i interval[d|m|w]] [-j] [-J journal_options]\n"
+		  "\t[-l] [-s sparse_flag] [-m reserved_blocks_percent]\n"
+		  "\t[-o [^]mount_options[,...]] [-r reserved_blocks_count]\n"
+		  "\t[-u user] [-C mount_count] [-L volume_label] "
+		  "[-M last_mounted_dir]\n"
+		  "\t[-O [^]feature[,...]] [-T last_check_time] [-U UUID]"
 		  " device\n"), program_name);
 	exit (1);
 }
@@ -621,7 +622,7 @@ static void parse_tune2fs_options(int argc, char **argv)
 					EXT2_FLAG_JOURNAL_DEV_OK;
 				break;
 			case 'm':
-				reserved_ratio = strtoul (optarg, &tmp, 0);
+				reserved_ratio = strtod(optarg, &tmp);
 				if (*tmp || reserved_ratio > 50) {
 					com_err (program_name, 0,
 						 _("bad reserved block ratio - %s"),
@@ -819,13 +820,12 @@ int main (int argc, char ** argv)
 	if (i_flag) {
 		sb->s_checkinterval = interval;
 		ext2fs_mark_super_dirty(fs);
-		printf (_("Setting interval between check %lu seconds\n"), interval);
+		printf (_("Setting interval between checks to %lu seconds\n"), interval);
 	}
 	if (m_flag) {
-		sb->s_r_blocks_count = (sb->s_blocks_count / 100)
-			* reserved_ratio;
+		sb->s_r_blocks_count = sb->s_blocks_count * reserved_ratio /100;
 		ext2fs_mark_super_dirty(fs);
-		printf (_("Setting reserved blocks percentage to %lu (%u blocks)\n"),
+		printf (_("Setting reserved blocks percentage to %g%% (%u blocks)\n"),
 			reserved_ratio, sb->s_r_blocks_count);
 	}
 	if (r_flag) {
